{
@abstract(provides standard operations on arrays and strings)
@author(Marco Schmidt (marcoschmidt@geocities.com))
@created(31 Jul 1998)
@lastmod(10 Apr 2000)

Perform basic operations on arrays of primitive data types: sort, convert
characters to another case, search - should be optimized with assembler
whenever possible.
}

unit arrays;

interface

{$I platform.inc}

uses
  numbers;

type
	{ array of characters }
	TCharArray = Array[0..30000] of Char; { will be overindexed }
	{ pointer to @link(TCharArray) }
	PCharArray = ^TCharArray;

{ Compares two arrays of byte, Source and Dest, both being N bytes
  large. Returns relation as -1 (Source < Dest), 0 (equal) or 1
  (Source > Dest). }
function CompareBytes(var Source, Dest; N: TSInt32): Integer;
{ Compares two arrays of characters, Source and Dest, both being N characters
  large. Before comparing two single characters, they're converted to
  uppercase, so 'hallo' and 'HalLo' will be considered equal.
  Returns relation as -1, 0 or 1 (meaning as in @link(CompareBytes)). }
function CompareCharsCI(var Source, Dest; N: TSInt32): Integer;
{ Compares two strings, Source and Dest, not regarding the case ('hallo' and
  'HallO' will be considered equal). Returns relation as -1, 0 or 1 (meaning
  as in @link(CompareBytes)). }
function CompareStringsCI(var Source, Dest: String): Integer;
{ Converts array of bytes Source to array of characters Dest, which will
  store a hexadecimal representation of the input bytes (example: 0 10 255
  will be converted to "000aff". Input array size is N bytes, output array
  must have space for 2 * N characters! }
procedure ConvertBytesToHex(var Source, Dest; N: TSInt32);
{ Copies character array Source to character array Dest, both N characters
  large, converting to lower case. }
procedure ConvertBytesToLowerCase(var Source, Dest; N: TSInt32);
{ Copies character array Source to character array Dest, both N characters
  large, converting to upper case. }
procedure ConvertBytesToUpperCase(var Source, Dest; N: TSInt32);
{ Converts array of characters Source, size 2 * N characters, to array of
  bytes Dest, N bytes large. It is assumed Source's characters are in
  hexadecimal format. Example: "0aff13" (six characters) will be converted to
  10 255 19 (three bytes). Does not check if input data is correct,
  non-hexadecimal input will result in zero output bytes. }
procedure ConvertHexToBytes(var Source, Dest; N: TSInt32);
{procedure FillBytes(var Data; N: TSInt32; B: Byte);}
{function FindByte(var Data; N: TSInt32; B: Byte): TSInt32;}
{ Searches array of bytes Data, size N, for minimum and maximum value.
  Returns these in Min and Max. }
procedure FindMinMaxBytes(var Data; N: TSInt32; var Min, Max: Byte);
{ Copies array of bytes Source to Dest, both N bytes large, taking each
  source value as an index into the array[0..255] of byte NewBytes, writing
  the looked-up value to Dest. }
procedure ReplaceBytes(var Source, Dest; N: TSInt32; var NewBytes);
{$ifndef PPC_DELPHI}
procedure SetLength(var S: String; L: Integer);
{$endif}
{procedure SortBytes(var Data; N: TSInt32);}
{ Sorts array of TNativeInt Data, N integers large, in ascending order,
  using a standard Quicksort implementation. }
procedure SortNativeInts(var Data; N: TSInt32);
{ Copies input string S to U, converting all characters to uppercase. }
procedure StringToUpper(var S, U: String);

implementation

const
	ToHexTable: Array[0..15] of Byte =
		(48, 49, 50, 51, 52, 53, 54, 55,
		 56, 57, 65, 66, 67, 68, 69, 70);

var
	FromHexTable: Array[0..255] of Byte;
	ToLowerTable: Array[0..255] of Byte;
	ToUpperTable: Array[0..255] of Byte;

function CompareBytes(var Source, Dest; N: TSInt32): Integer;
var
	I: TSInt32;
begin
	CompareBytes := 0;
	I := 0;
	while (I < N) do
	begin
		if (TByteArray(Source)[I] < TByteArray(Dest)[I])
		then
			begin
				CompareBytes := -1;
				exit;
			end
		else
			if (TByteArray(Source)[I] > TByteArray(Dest)[I])
			then
				begin
					CompareBytes := 1;
					exit;
				end;
		Inc(I);
	end;
end;

function CompareCharsCI(var Source, Dest; N: TSInt32): Integer;
var
	I: LongInt;
	C1: Integer;
	C2: Integer;
begin
	I := 0;
	while (I < N) do
	begin
		C1 := ToUpperTable[TByteArray(Source)[I]];
		C2 := ToUpperTable[TByteArray(Dest)[I]];
		if (C1 <> C2)
		then
			begin
				if (C1 < C2)
				then CompareCharsCI := -1
				else CompareCharsCI := 1;
				exit;
			end;
		Inc(I);
	end;
	CompareCharsCI := 0;
end;

function CompareStringsCI(var Source, Dest: String): Integer;
var
	L1: Byte;
	L2: Byte;
	Minimum: Byte;
	R: TSInt32;
	Rel: Integer;
begin
	L1 := Length(Source);
	L2 := Length(Dest);
	if (L1 < L2)
	then
		begin
			Rel := -1;
			Minimum := L1
		end
	else
	if (L1 > L2)
	then
		begin
			Rel := 1;
			Minimum := L2;
		end
	else
		begin
			Rel := 0;
			Minimum := L1;
		end;
	R := CompareCharsCI(Source[1], Dest[1], Minimum);
	if (R = 0)
	then CompareStringsCI := Rel
	else CompareStringsCI := R;
end;

{
N = number of bytes to be converted
Source must be N bytes large
Dest must be N * 2 bytes large
Source must not be the same as Dest
}
procedure ConvertBytesToHex(var Source, Dest; N: TSInt32);
var
	B: Byte;
	I: TSInt32;
	J: TSInt32;
begin
	I := 0;
	J := 0;
	while (I <> N) do
	begin
		B := TByteArray(Source)[I];
		Inc(I);		
		TByteArray(Dest)[J] := ToHexTable[B shr 4];
		Inc(J);
		TByteArray(Dest)[J] := ToHexTable[B and 15];
		Inc(J);
	end;
end;

procedure ConvertBytesToLowerCase(var Source, Dest; N: TSInt32);
begin
	ReplaceBytes(Source, Dest, N, ToLowerTable);
end;

procedure ConvertBytesToUpperCase(var Source, Dest; N: TSInt32);
begin
	ReplaceBytes(Source, Dest, N, ToUpperTable);
end;

procedure ConvertHexToBytes(var Source, Dest; N: TSInt32);
var
	I: TSInt32;
	J: TSInt32;
begin
	I := 0;
	J := 0;
	while (J <> N) do
	begin
		TByteArray(Dest)[J] :=
			(FromHexTable[TByteArray(Source)[I]] shr 4) or
			FromHexTable[TByteArray(Source)[I + 1]];
		Inc(J);
		Inc(I, 2);
	end;
end;

procedure FindMinMaxBytes(var Data; N: TSInt32; var Min, Max: Byte);
var
	B: Byte;
	I: TSInt32;
begin
	Max := TByteArray(Data)[0];
	Min := TByteArray(Data)[0];
	I := 1;
	while (I < N) do
	begin
		B := TByteArray(Data)[I];
		if (B < Min)
		then Min := B;
		if (B > Max)
		then Max := B;
		Inc(I);
	end;
end;

procedure InitTables;
var
	I: Integer;
begin
	for I := 0 to 255 do
	begin
		FromHexTable[I] := 0;
		if (I >= 65) and (I <= 90) { A .. Z }
		then
			begin
				ToLowerTable[I] := I + 32;
				FromHexTable[I] := I - 55;
			end
		else ToLowerTable[I] := I;
		if (I >= 97) and (I <= 122) { a .. z }
		then
			begin
				ToUpperTable[I] := I - 32;
				FromHexTable[I] := I - 87;
			end
		else ToUpperTable[I] := I;
		if (I >= 48) and (I <= 57) { 0 .. 9 }
		then FromHexTable[I] := I - 48;
	end;
end;

procedure ReplaceBytes(var Source, Dest; N: TSInt32; var NewBytes);
var
	I: TSInt32;
begin
	I := 0;
	while (I <> N) do
	begin
		TByteArray(Dest)[I] := TByteArray(NewBytes)[TByteArray(Source)[I]];
		Inc(I);
	end;
end;

{$ifndef PPC_DELPHI}
procedure SetLength(var S: String; L: Integer);
begin
	S[0] := Chr(L);
end;
{$endif}

procedure SortNativeInts(var Data; N: TSInt32);
type
	TNativeIntArray = Array[0..0] of TNativeInt;

	procedure SubSort(L, R: TSInt32);
	var
		I, J: TSInt32;
		X, Y: TNativeInt;
	begin
		I := L;
		J := R;
		X := TNativeIntArray(Data)[(L + R) div 2];
		repeat
			while (TNativeIntArray(Data)[I] < X) do
				I := I + 1;
			while (X < TNativeIntArray(Data)[J]) do
				J := J - 1;
			if (I <= J)
			then
				begin
					Y := TNativeIntArray(Data)[I];
					TNativeIntArray(Data)[I] := TNativeIntArray(Data)[J];
					TNativeIntArray(Data)[J] := Y;
					I := I + 1;
					J := J - 1;
				end;
		until (I > J);
		if (L < J)
		then SubSort(L, J);
		if (I < R)
		then SubSort(I, R);
	end;

begin
	if (N > 1)
	then SubSort(0, N - 1);
end;

procedure StringToUpper(var S, U: String);
begin
	ConvertBytesToUppercase(S[1], U[1], Length(S));
	SetLength(U, Length(S));
end;

begin
	InitTables;
end.