{
@abstract(provides basic classes for input and output streams)
@author(Marco Schmidt (marcoschmidt@geocities.com))
@created(7 May 1998)
@lastmod(14 Aug 1999)

Text files to be read by TInputStream should stick to one of the usual
line ending formats: DOS (13 10), Unix (10) or Mac (13). Anything else,
like (13 13 10) will make counting lines impossible (though this sort
of formatting is used in some word processors).

When reading / writing raw data, function return values tell about how much
data has actually been read / written. If there were errors, -1 will be
returned. New derived stream types must keep this system.
}
unit streams;

interface

uses
  numbers;

type
  { integer type for stream offsets / lengths; must be signed so that we can return -1 on errors }
  TStreamOffset = TSInt32;
  { supported text line types }
  TTextType = (TEXT_TYPE_DOS, TEXT_TYPE_UNIX, TEXT_TYPE_MAC);
  { basic class for input and output streams }
  TStream = object
    { is true if end of stream has been reached, false otherwise }
    EndOfStream: Boolean;
    { Initializes all fields }
    constructor Init;
    { First calls @link(Close), then @link(ReleaseBuffer). }
    destructor Done; virtual;
    { Does nothing - heirs will override this procedure and actually implement some functionality. }
    procedure Close; virtual;
    { Returns if an error has occurred. }
    function Error: Boolean;
    { Returns the current size of the buffer. }
    function GetBufferSize: TUInt32;
    { Returns the current byte order. }
    function GetByteOrder: TByteOrder;
    { Returns the current line number. This will only make sense if
      the stream is a text stream which is read / written by
      @link(TInputStream.ReadLine) / @link(TOutputStream.WriteLine). }
    function GetLineNumber: TUInt32;
        function GetName: String;
    { Returns current position in stream or -1 if there were errors. }
    function GetPos: TStreamOffset;
    { Returns current size of stream or -1 if there were errors. }
    function GetSize: TStreamOffset;
    { Returns current text type, i.e. the sort of lines we are reading /
      writing. Supported are Dos, Unix and Mac types. }
    function GetTextType: TTextType;
    { Returns true if we have a buffer allocated. }
    function IsBuffered: Boolean;
        function IsOpen: Boolean;
    { Releases the buffer if one has been allocated. }
    procedure ReleaseBuffer;
    { Resets the size of the buffer. All buffer content will be lost. }
    procedure SetBufferSize(NewBufferSize: TSInt32);
    { Sets a new byte order. }
    procedure SetByteOrder(NewByteOrder: TByteOrder);
        procedure SetName(N: String);
        procedure SetOpenStatus(IsNowOpen: Boolean);
    { Sets current position in file to argument value. }
    procedure SetPos(const NewPosition: TStreamOffset);
    { Sets a new text type. }
    procedure SetTextType(NewTextType: TTextType);
  private
    { array of bytes used to speed up read / write access }
    Buffer: PByteArray;
    { number of bytes currently in buffer }
    BufferContent: TSInt32;
    { }
    BufferIndex: TSInt32;
    { number of bytes allocated to @link(Buffer) }
    BufferSize: TSInt32;

    ByteOrder: TByteOrder;
    { }
    ErrorCode: TUInt32;
        {}
        HasBeenOpened: Boolean;
    { current line number, can be determined via @link(GetLineNumber) }
    LineNumber: TUInt32;
        MyName: String;
    { determines if random access to this stream is possible }
    RandomAccess: Boolean;
    { current text type, can be accessed via @link(GetTextType) and @link(SetTextType) }
    TextType: TTextType;
  end;
  { pointer to @link(TInputStream) }
  PInputStream = ^TInputStream;
  { basic input stream class }
  TInputStream = object(TStream)
    function CheckBytes(N: TStreamOffset; var Data): Boolean;
    function EOS: Boolean; { end of stream }
    procedure FillBuffer;
        function HasData: Boolean; virtual;
    function PeekByte: TSInt32;
    function ReadByte: TSInt32;
    function ReadBytes(N: TStreamOffset; var Dest): TStreamOffset;
    procedure ReadChar(var C: Char);
    procedure ReadLine(var S: String);
    function ReadRawBytes(N: TStreamOffset; var Dest): TStreamOffset; virtual;
    procedure ReadSInt8(var Value: TSInt8);
    procedure ReadSInt16(var Value: TSInt16);
    procedure ReadSInt32(var Value: TSInt32);
    procedure ReadUInt8(var Value: TUInt8);
    procedure ReadUInt16(var Value: TUInt16);
    procedure ReadUInt32(var Value: TUInt32);
    procedure Skip(N: TStreamOffset); virtual;
  end;
  { pointer to @link(TOutputStream) }
  POutputStream = ^TOutputStream;
  { basic output stream class }
  TOutputStream = object(TStream)
    procedure Close; virtual;
    procedure Flush;
    procedure Truncate; virtual;
    procedure WriteByte(var Value: Byte);
    function WriteBytes(N: TStreamOffset; const Source): TStreamOffset;
    procedure WriteChar(const C: Char);
    procedure WriteLine(const S: String);
    function WriteRawBytes(N: TStreamOffset; var Source): TStreamOffset; virtual;
    procedure WriteSInt8(var Value: TSInt8);
    procedure WriteSInt16(var Value: TSInt16);
    procedure WriteSInt32(var Value: TSInt32);
    procedure WriteString(const S: String);
    procedure WriteUInt8(var Value: TUInt8);
    procedure WriteUInt16(var Value: TUInt16);
    procedure WriteUInt32(var Value: TUInt32);
  end;

{$I platform.inc}

const
  { buffer size to be used if not explicitly set by caller via @link(TStream.SetBufferSize) }
  DEFAULT_BUFFER_SIZE: TUInt32 = 4096; { 4 KB }
    { default text type to be used, platform-dependent }
  DEFAULT_TEXT_TYPE: TTextType =
  {$ifdef OS_LINUX}
    TEXT_TYPE_UNIX;
  {$else}
    {$ifdef OS_MAC}
      TEXT_TYPE_MAC;
    {$else}
      TEXT_TYPE_DOS;
    {$endif}
  {$endif}
  { minimum possible buffer size in combination with @link(TStream.SetBufferSize) }
  MINIMUM_BUFFER_SIZE: TSInt32 = 16;

implementation

uses
  arrays;

const
  { strings representing line endings for all text types }
  TLineEndings: Array[TEXT_TYPE_DOS..TEXT_TYPE_MAC] of String[2] =
    (#13#10, #10, #13);

{ TStream }

constructor TStream.Init;
begin
  EndOfStream := false;
  ErrorCode := 0;
    HasBeenOpened := false;
  LineNumber := 0;
  RandomAccess := false;
  Buffer := nil;
  { now that Buffer is nil, ReleaseBuffer simply clears all buffer variables }
  ReleaseBuffer;
  { TO DO: change the following calls according to home platform!  }
  SetByteOrder(BYTE_ORDER_INTEL);
  SetTextType(DEFAULT_TEXT_TYPE);
end;

destructor TStream.Done;
begin
  Close;
  ReleaseBuffer;
end;

procedure TStream.Close;
begin
  HasBeenOpened := false;
end;

function TStream.Error: Boolean;
begin
  Error := (ErrorCode <> 0);
end;

function TStream.GetBufferSize: TUInt32;
begin
  GetBufferSize := BufferSize;
end;

function TStream.GetByteOrder: TByteOrder;
begin
  GetByteOrder := ByteOrder;
end;

function TStream.GetLineNumber: TUInt32;
begin
  GetLineNumber := LineNumber;
end;

function TStream.GetName: String;
begin
  GetName := MyName;
end;

function TStream.GetPos: TStreamOffset;
begin
  { if descendant stream type allows it, overwrite this! }
  GetPos := -1; { meaning: we don't know }
end;

function TStream.GetSize: TStreamOffset;
begin
  { if descendant stream type allows it, overwrite this! }
  GetSize := -1; { meaning: we don't know }
end;

function TStream.GetTextType: TTextType;
begin
  GetTextType := TextType;
end;

function TStream.IsBuffered: Boolean;
begin
  IsBuffered := Assigned(Buffer);
end;

function TStream.IsOpen: Boolean;
begin
  IsOpen := HasBeenOpened;
end;

procedure TStream.ReleaseBuffer;
begin
  if (Buffer <> nil)
  then FreeMem(Buffer, BufferSize);
  Buffer := nil;
  BufferContent := 0;
  BufferIndex := 0;
  BufferSize := 0;
end;

procedure TStream.SetBufferSize(NewBufferSize: TSInt32);
begin
  if (NewBufferSize < MINIMUM_BUFFER_SIZE) or
      (NewBufferSize = BufferSize)
  then exit;
  ReleaseBuffer;
  if (NewBufferSize <> 0)
  then
    begin
      GetMem(Buffer, NewBufferSize);
      if Assigned(Buffer)
      then BufferSize := NewBufferSize;
    end;
end;

procedure TStream.SetByteOrder(NewByteOrder: TByteOrder);
begin
  ByteOrder := NewByteOrder;
end;

procedure TStream.SetName(N: String);
begin
  MyName := N;
end;

procedure TStream.SetOpenStatus(IsNowOpen: Boolean);
begin
  HasBeenOpened := IsNowOpen;
end;

procedure TStream.SetPos(const NewPosition: TStreamOffset);
begin
  { abstract }
end;

procedure TStream.SetTextType(NewTextType: TTextType);
begin
  TextType := NewTextType;
end;

{ TInputStream }

function TInputStream.CheckBytes(N: TStreamOffset; var Data): Boolean;
var
  I: TStreamOffset;
  R: TSInt32;
begin
  I := 0;
  while (I <> N) do
  begin
    R := ReadByte;
    if (R = -1) or (R <> TByteArray(Data)[I])
    then 
      begin
        CheckBytes := false;
        exit;
      end;
    Inc(I);
  end;
  CheckBytes := true;
end;

function TInputStream.EOS: Boolean;
begin
  EOS := EndOfStream;
end;

procedure TInputStream.FillBuffer;
begin
  if (not Assigned(Buffer))
  then
    begin
      SetBufferSize(DEFAULT_BUFFER_SIZE);
      if (not Assigned(Buffer))
      then exit;
    end;
  BufferContent := ReadRawBytes(BufferSize, Buffer^);
  BufferIndex := 0;
end;

function TInputStream.HasData: Boolean;
begin
  HasData := (BufferIndex < BufferContent);
end;

function TInputStream.PeekByte: TSInt32;
begin
  if (BufferIndex >= BufferContent)
  then
    begin
      FillBuffer;
      if (BufferContent <= 0)
      then PeekByte := -1
      else
        begin
          PeekByte := Buffer^[0];
          BufferIndex := 0;
        end;
    end
  else
    begin
      PeekByte := Buffer^[BufferIndex];
    end;
end;

{ same functionality as ReadUInt8; but this one is faster, as it does not
  use ReadBytes (as ReadUInt8 does) }
function TInputStream.ReadByte: TSInt32;
begin
  if (BufferIndex >= BufferContent)
  then
    begin
      FillBuffer;
      if (BufferContent <= 0)
      then ReadByte := -1
      else
        begin
          ReadByte := Buffer^[0];
          BufferIndex := 1;
        end;
    end
  else
    begin
      ReadByte := Buffer^[BufferIndex];
      Inc(BufferIndex);
    end;
end;

function TInputStream.ReadBytes(N: TStreamOffset; var Dest): TStreamOffset;
var
  I: TStreamOffset;
  B: TStreamOffset;
  SaveN: TStreamOffset;
begin
  SaveN := N;
  I := 0; { index into TByteArray(Dest) }
  while (N > 0) do
  begin
    if (BufferIndex >= BufferContent)
    then
      begin
        FillBuffer;
        if (BufferContent <= 0)
        then
          begin
            ReadBytes := -1;
            exit;
          end;
      end
    else
      begin
        { compute number of bytes left in buffer }
        B := BufferContent - BufferIndex;
        { if less byte to get than in buffer, set B to N }
        if (N < B)
        then B := N;
        { copy from buffer to destination }
        System.Move(Buffer^[BufferIndex], TByteArray(Dest)[I], B);
        { }
        Inc(BufferIndex, B);
        Dec(N, B);
        Inc(I, B);
      end;
  end;
  ReadBytes := SaveN;
end;

{
Reads a single character from input, regarding SizeOf(Char).
Does not deal with byte order in case SizeOf(Char) > 2.
}
procedure TInputStream.ReadChar(var C: Char);
begin
  ReadBytes(SizeOf(Char), C);
end;

{
Reads a text line from input (a complete line is read in, whether it fits
into S or not - all addtional characters will be lost.
However, it is checked how many characters S can hold, so no memory
behind S is overwritten.
Relies on Char being one byte large.
}
procedure TInputStream.ReadLine(var S: String);
var
  C: Char;
  I: TUInt32;
  StringSize: TUInt32;
begin
  I := 0;
  StringSize := SizeOf(S) - 1;
  repeat
    ReadChar(C);
    if (I < StringSize)
      then
        begin
          Inc(I);
          S[I] := C;
        end
      else
        begin
          Inc(I);
        end;
  until (C = #10) or (C = #13) or EOS;
  if (C = #13)
    then
      begin
        if (PeekByte =   10)
          then ReadChar(C);
      end;
  SetLength(S, I - 1);
  { S[0] := Chr(I - 1); }
  Inc(LineNumber);
end;

function TInputStream.ReadRawBytes(N: TStreamOffset; var Dest): TStreamOffset;
begin
  ReadRawBytes := -1;
end;

procedure TInputStream.ReadSInt8(var Value: TSInt8);
begin
  ReadBytes(SizeOf(TSInt8), Value);
end;

procedure TInputStream.ReadSInt16(var Value: TSInt16);
var
  RawData: Array[0..1] of Byte;
begin
  ReadBytes(2, RawData);
  GetSInt16(ByteOrder, RawData, Value);
end;

procedure TInputStream.ReadSInt32(var Value: TSInt32);
var
  RawData: Array[0..3] of Byte;
begin
  ReadBytes(4, RawData);
  GetSInt32(ByteOrder, RawData, Value);
end;

procedure TInputStream.ReadUInt8(var Value: TUInt8);
begin
  ReadBytes(SizeOf(TUInt8), Value);
end;

procedure TInputStream.ReadUInt16(var Value: TUInt16);
var
  RawData: Array[0..1] of Byte;
begin
  ReadBytes(2, RawData);
  GetUInt16(ByteOrder, RawData, Value);
end;

procedure TInputStream.ReadUInt32(var Value: TUInt32);
var
  RawData: Array[0..3] of Byte;
begin
  ReadBytes(4, RawData);
  GetUInt32(ByteOrder, RawData, Value);
end;

procedure TInputStream.Skip(N: TStreamOffset);
begin
  if Assigned(Buffer)
  then
    begin
      while (N > 0) do
      begin
        ReadByte;
        Dec(N);
      end;
    end
  else
    begin
      while (N > 0) do
      begin
        ReadByte;
        Dec(N);
      end;
    end;
end;

{ TOutputStream }

procedure TOutputStream.Close;
begin
    inherited Close;
  Flush;
end;

procedure TOutputStream.Flush;
begin
  if Assigned(Buffer) and (BufferContent > 0)
  then
    begin
      if (WriteRawBytes(BufferContent, Buffer^) <> BufferContent)
      then ;
      BufferContent := 0;
    end;
end;

procedure TOutputStream.Truncate;
begin
  { abstract }
end;

procedure TOutputStream.WriteByte(var Value: Byte);
begin
  if Assigned(Buffer)
  then
    begin
      Buffer^[BufferContent] := Value;
      Inc(BufferContent);
      if (BufferContent = BufferSize)
      then Flush;
    end
  else
    begin
      SetBufferSize(DEFAULT_BUFFER_SIZE);
      if Assigned(Buffer)
      then WriteByte(Value);
    end;
end;

function TOutputStream.WriteBytes(N: TStreamOffset; const Source): TStreamOffset;
var
  Bytes: TStreamOffset;
  FreeBytes: TStreamOffset;
  I: TStreamOffset;
  SaveN: TStreamOffset;
begin
  I := 0;
  SaveN := N;
  while (N > 0) do
  begin
    if Assigned(Buffer)
    then
      begin
        FreeBytes := BufferSize - BufferContent;
        if (FreeBytes > N)
        then Bytes := N
        else Bytes := FreeBytes;
        System.Move(TByteArray(Source)[I], Buffer^[BufferContent], Bytes);
        Inc(BufferContent, Bytes);
        Inc(I, Bytes);
        Dec(N, Bytes);
        if (BufferContent = BufferSize)
        then Flush;
      end
    else
      begin
        SetBufferSize(DEFAULT_BUFFER_SIZE);
        if (not Assigned(Buffer))
        then
          begin
            WriteBytes := -1;
            exit;
          end;
      end;
  end;
  WriteBytes := SaveN;
end;

procedure TOutputStream.WriteChar(const C: Char);
begin
  WriteBytes(SizeOf(Char), C);
end;

procedure TOutputStream.WriteLine(const S: String);
begin
  WriteBytes(Length(S) * SizeOf(Char), S[1]);
  WriteBytes(Length(TLineEndings[TextType]), TLineEndings[TextType][1]);
end;

function TOutputStream.WriteRawBytes(N: TStreamOffset; var Source): TStreamOffset;
begin
  WriteRawBytes := -1;
end;

procedure TOutputStream.WriteSInt8(var Value: TSInt8);
begin
   WriteByte(Byte(Value));
end;

procedure TOutputStream.WriteSInt16(var Value: TSInt16);
var
  I: TSInt16;
begin
  SetSInt16(ByteOrder, Value, I);
  WriteBytes(2, I);
end;

procedure TOutputStream.WriteSInt32(var Value: TSInt32);
var
  I: TSInt32;
begin
  SetSInt32(ByteOrder, Value, I);
  WriteBytes(4, I);
end;

procedure TOutputStream.WriteString(const S: String);
var
  I: Word;
begin
  if (Length(S) = 0)
  then exit;
  for I := 1 to Length(S) do
    WriteChar(S[I]);
end;

procedure TOutputStream.WriteUInt8(var Value: TUInt8);
begin
  WriteByte(Value);
end;

procedure TOutputStream.WriteUInt16(var Value: TUInt16);
var
  I: TUInt16;
begin
  SetUInt16(ByteOrder, Value, I);
  WriteBytes(2, I);
end;

procedure TOutputStream.WriteUInt32(var Value: TUInt32);
var
  I: TUInt32;
begin
  SetUInt32(ByteOrder, Value, I);
  WriteBytes(4, I);
end;

begin
end.